import javax.microedition.lcdui.*;
import java.util.*;

public class BrowserCanvas extends Canvas implements CommandListener, Receiver {
    static final int BLINK_PERIOD = 1000;
    static final int REPEAT_DELAY = 1000;
    static final int REPEAT_PERIOD = 100;

    static final int BACKGROUND_COLOR = 0xffffff;
    static final int SELECTION_COLOR = 0x800080;
    static final int FOCUS_COLOR = 0x000000;
    static final int EPSILON = 10;

    int        x, y;
    int        focusX, focusY;
    int        focusWidth, focusHeight;
    int        focusX0, focusY0;
    int        screenWidth, screenHeight;
    int        selectionColor = SELECTION_COLOR;
    MapBrowser browser;
    int        level;
    int        scale;
    int        zoom;
    int        imageSize;
    boolean    busy;
    int        skipImages;
    Image[]    imageCache;
    String[]   imageUrl;
    int        imageCachePtr;
    int        stepX, stepY;
    int        focusStepX, focusStepY;
    Timer      repeatTimer = new Timer();
    Timer      blinkTimer = new Timer();
    RepeatTask repeatTask;
    BlinkTask  blinkTask;
    int        cacheSize;
    Location[] selection;
    Index      selectionIndex;
    int        currSelection;

    String buildImageUrl(int x, int y) { 
        return browser.url + level + '/' + browser.city + y + '_' + x + ".png";
    }

    Image getImage(int x, int y) { 
        String url = buildImageUrl(x, y);
        for (int i = 0; i < cacheSize; i++) {
            if (url.equals(imageUrl[i])) { 
                return imageCache[i];
            }
        }
        return null;
    }

    void startImageLoading(int x, int y) { 
        busy = true;
        new InternetConnection(buildImageUrl(x, y), this);
    }
    
    public void failure(String url) { 
        System.out.println("Failed to load image " + url);
        // try once again
        new InternetConnection(url, this);
    }
    
    public void received(String url, byte[] data, int len) { 
        // System.out.println("Receive image " + url);
        Image image = Image.createImage(data, 0, len);        
        imageCache[imageCachePtr] = image;
        imageUrl[imageCachePtr] = url;
        imageCachePtr = (imageCachePtr + 1) % cacheSize;
        repaint();
    }

    void setPositions(Index index, Location[] locations) { 
        selection = locations;
        selectionIndex = index;
        currSelection = -1;
        if (locations.length == 1) {
            Location l = locations[0];
            if ((browser.settings.flags & Settings.AUTO_ZOOM) != 0) { 
                for (scale = 1, level = browser.levels; 
                     (l.getWidth() > screenWidth*scale || l.getHeight() > screenHeight*scale)
                      && level > 1;                
                     level -= 1, scale *= zoom);
                if (level != 1) { 
                    removeCommand(MapBrowser.ZOOM_OUT_CMD);
                } 
                if (level != browser.levels) { 
                    addCommand(MapBrowser.ZOOM_IN_CMD);
                }
                setFocus();
            }
            x = l.getCenterX() - screenWidth*scale/2;
            y = l.getCenterY() - screenHeight*scale/2;
        }
        if (locations.length > 0) { 
            if (cacheSize >= 4 && (browser.settings.flags & Settings.BLINKING) != 0) { 
                startBlinking();
            }
            if (index.type == Index.POINT) { 
                addCommand(MapBrowser.NEXT_CMD);
            }
            addCommand(MapBrowser.CLEAN_CMD);
        }
        check();
    }

    void startBlinking() { 
        blinkTask = new BlinkTask();
        blinkTimer.schedule(blinkTask, BLINK_PERIOD, BLINK_PERIOD);
    }

    void stopBlinking() { 
        if (blinkTask != null) { 
            blinkTask.cancel();
            blinkTask = null;
        }
    }

    void setFocus() { 
        focusX0 = screenWidth*(scale-1)/2;
        focusY0 = screenHeight*(scale-1)/2;
        focusX = x + focusX0;
        focusY = y + focusY0;        
    }        

    BrowserCanvas(MapBrowser browser) { 
        setFullScreenMode(true);
        this.browser = browser;
        setCommandListener(this);
        addCommand(MapBrowser.INDEX_CMD);
        addCommand(MapBrowser.SETTINGS_CMD);
        addCommand(MapBrowser.HELP_CMD);
        addCommand(MapBrowser.ABOUT_CMD);
        addCommand(MapBrowser.QUIT_CMD);
        cacheSize = browser.settings.cacheSize;
        imageCache = new Image[cacheSize];
        imageUrl = new String[cacheSize];
        screenWidth = getWidth();
        screenHeight = getHeight();
        stepX = screenWidth/4;
        stepY = screenHeight/4;
        zoom = browser.zoom;
        scale = 1;
        x = 0;
        y = 0;
        level = 1;
        focusWidth = screenWidth/zoom;
        focusHeight = screenHeight/zoom;
        focusStepX = focusWidth/zoom;
        focusStepY = focusHeight/zoom;

        for (int i = browser.levels; --i != 0;) { 
            scale *= zoom;
        }
        if (browser.levels > 1) { 
            addCommand(MapBrowser.ZOOM_IN_CMD);
        }
        imageSize = browser.cell;
        setFocus();
        check();
        Display.getDisplay(browser).setCurrent(this);
    }
        
    protected void paint(Graphics g) {
        if (skipImages == 0 && level == 1) { 
            g.setColor(BACKGROUND_COLOR);
            g.fillRect(g.getClipX(), g.getClipY(), g.getClipWidth(), g.getClipHeight());
        }
        //System.out.println("Repaint xc=" + g.getClipX() + ", yc=" + g.getClipY() + ", width=" + g.getClipWidth() + ", height=" + g.getClipHeight() + ", x=" + x + ", y=" + y + ", scale=" + scale);
        //System.out.println("Draw x=" + x + ", y=" + y + ", focusX = " + focusX + ", focusY=" + focusY + ", scale=" + scale);
        int x0 = x/scale + g.getClipX();
        int y0 = y/scale + g.getClipY();
        int x1 = x0 + g.getClipWidth() - 1;
        int y1 = y0 + g.getClipHeight() - 1;
        if (x1 >= browser.width/scale) { 
            x1 = browser.width/scale-1;
        }
        if (y1 >= browser.width/scale) { 
            y1 = browser.width/scale-1;
        }
        int n = 0;
        for (int i = x0/imageSize; i <= x1/imageSize; i++) { 
            for (int j = y0/imageSize; j <= y1/imageSize; j++) { 
                if (++n >= skipImages) { 
                    Image image = getImage(i, j);
                    if (image != null) { 
                        // System.out.println("Draw image j=" + j + ", i=" + i + " at x=" + (i*imageSize-x/scale) + ", y=" + (j*imageSize-y/scale));
                        g.drawImage(image, i*imageSize-x/scale, j*imageSize-y/scale, Graphics.TOP|Graphics.LEFT);
                    } else { 
                        if (n > skipImages) { 
                            skipImages = n;
                            startImageLoading(i, j);
                        }
                        return;
                    }
                }
            }
        }
        busy = false;
        skipImages = 0;

        if (level != browser.levels) { 
            g.setColor(FOCUS_COLOR);
            g.setStrokeStyle(Graphics.DOTTED);
            g.drawRect((focusX-x)/scale, (focusY-y)/scale, focusWidth, focusHeight);
        }
        if (selection != null) { 
            g.setStrokeStyle(Graphics.DOTTED);
            g.setColor(selectionColor);
            for (int i = 0; i < selection.length; i++) { 
                Location loc = selection[i];
                switch (selectionIndex.type) { 
                case Index.POINT:
                    g.drawLine((loc.points[0]-x)/scale-EPSILON, (loc.points[1]-y)/scale, 
                               (loc.points[0]-x)/scale+EPSILON, (loc.points[1]-y)/scale);
                    g.drawLine((loc.points[0]-x)/scale, (loc.points[1]-y)/scale-EPSILON, 
                               (loc.points[0]-x)/scale, (loc.points[1]-y)/scale+EPSILON);
                    if (i == currSelection) { 
                        g.drawArc((loc.points[0]-x)/scale-EPSILON, (loc.points[1]-y)/scale-EPSILON,
                                  EPSILON*2, EPSILON*2, 0, 360);
                    }
                    break;
                case Index.RECTANGLE:
                    g.drawRect((loc.points[0]-x)/scale, (loc.points[1]-y)/scale, 
                               (loc.points[2]-loc.points[0])/scale+1, 
                               (loc.points[3]-loc.points[1])/scale+1);
                    break;
                case Index.LINE:
                    for (int j = 2; j < loc.points.length; j += 2) { 
                        g.drawLine((loc.points[j-2]-x)/scale, (loc.points[j-1]-y)/scale, 
                                   (loc.points[j]-x)/scale, (loc.points[j+1]-y)/scale);
                    }
                    break;
                }
            }
        }
    }    

    boolean zoomIn() { 
        if (scale == 1) { 
            return false;
        }
        level += 1;
        scale /= zoom;
        x = focusX;
        y = focusY;
        addCommand(MapBrowser.ZOOM_OUT_CMD);
        if (scale == 1) { 
            removeCommand(MapBrowser.ZOOM_IN_CMD);
        } else { 
            setFocus();
            check();
        }
        return true;
    }        

    boolean zoomOut() { 
        if (level == 1) { 
            return false;
        }
        scale *= zoom;
        level -= 1;
        focusX = x;
        focusY = y;
        focusX0 = screenWidth*(scale-1)/2;
        focusY0 = screenHeight*(scale-1)/2;
        x -= focusX0;
        y -= focusY0;
        addCommand(MapBrowser.ZOOM_IN_CMD);
        if (level == 1) { 
            removeCommand(MapBrowser.ZOOM_OUT_CMD);
        } 
        return true;
    }        

    public void commandAction(Command c, Displayable d)  
    {
        if (c == MapBrowser.QUIT_CMD) {
            browser.quit();
        } else if (c == MapBrowser.ABOUT_CMD) {
            new HelpForm(browser, Locale.current.getResource("About"), Locale.current.getResource("AboutText"), this);
        } else if (c == MapBrowser.HELP_CMD) {
            new HelpForm(browser, Locale.current.getResource("Help"), Locale.current.getResource("HelpText"), this);
        } else if (c == MapBrowser.ZOOM_IN_CMD) {
            zoomIn();
            check();
        } else if (c == MapBrowser.ZOOM_OUT_CMD) {
            zoomOut();
            check();
        } else if (c == MapBrowser.INDEX_CMD) {
            new GroupList(this);
        } else if (c == MapBrowser.NEXT_CMD) {
            nextPoint();
            repaint();
        } else if (c == MapBrowser.INFO_CMD) {
            pointInfo();
        } else if (c == MapBrowser.CLEAN_CMD) {
            selection = null;
            stopBlinking();
            removeCommand(MapBrowser.CLEAN_CMD);
            removeCommand(MapBrowser.NEXT_CMD);
            removeCommand(MapBrowser.INFO_CMD);            
        } else if (c == MapBrowser.SETTINGS_CMD) {
            new SettingsForm(this);            
        }
    }
 
    class RepeatTask extends TimerTask { 
        RepeatTask(int keyCode) { 
            this.keyCode = keyCode;
        }

        public void run() { 
            if (!busy) { 
                keyAction(keyCode);
            }
        }

        int keyCode;
    }

    class BlinkTask extends TimerTask { 
        public void run() { 
            if (!busy) { 
                selectionColor ^= 0xffffff;
                repaint();
            }
        }
    }

    public void pointerPressed(int x, int y) { 
        focusX = this.x + x*scale;
        focusY = this.y + y*scale;
        repaint();
    }

    public void pointerReleased(int x, int y) { 
        focusX = this.x + x*scale;
        focusY = this.y + y*scale;
        repaint();
    }

    public void pointerDragged(int x, int y) { 
        this.x = focusX - x*scale;
        this.y = focusY - y*scale;
        check();
        repaint();
    }

    public void keyPressed(int keyCode) {
        if (repeatTask != null) { 
            repeatTask.cancel();
        }
        if (!busy) { 
            repeatTask = new RepeatTask(keyCode);
            repeatTimer.schedule(repeatTask, REPEAT_DELAY, REPEAT_PERIOD);
        }
    }

    public void keyReleased(int keyCode) {
        if (!busy) { 
            keyAction(keyCode);
        }
        if (repeatTask != null) { 
            repeatTask.cancel();
            repeatTask = null;
        }
    }
    
    private boolean pointInfo() { 
        if (selection != null && currSelection >= 0) { 
            Location loc = selection[currSelection];
            new InfoForm(this, selectionIndex, selection[currSelection], this);
            return true;
        }        
        return false;
    }

    private boolean nextPoint() { 
        if (selection != null && selectionIndex.type == Index.POINT) { 
            for (int i = 0; i < selection.length; i++) { 
                if (++currSelection == selection.length) { 
                    currSelection = 0;
                }
                Location loc = selection[currSelection];
                if (loc.points[0] >= x && loc.points[0] - x < screenWidth*scale
                    && loc.points[1] >= y && loc.points[1] - y < screenHeight*scale)
                {
                    addCommand(MapBrowser.INFO_CMD);
                    return true;
                }
            }
            currSelection = -1;
        }
        return false;
    }

    private void keyAction(int keyCode) {
        int dx = 0, dy = 0;

	switch (keyCode) {
        case KEY_NUM0:
            if (!pointInfo()) { 
                browser.error(Locale.current.getResource("No current point"), this);
            }
            return;
        case KEY_NUM1:
            dx = -1;
            dy = -1;
            break;
        case KEY_NUM2:
            dy = -1;
            break;
        case KEY_NUM3:
            dx = 1;
            dy = -1;
            break;
        case KEY_NUM4:
            dx = -1;
            break;
        case KEY_NUM5:
            if (!nextPoint()) { 
                browser.error(Locale.current.getResource("No points selected"), this);
            }
            repaint();
            return;
        case KEY_NUM6:
            dx = 1;
            break;
        case KEY_NUM7:
            dx = -1;
            dy = 1;
            break;
        case KEY_NUM8:
            dy = 1;
            break;
        case KEY_NUM9:
            dx = 1;
            dy = 1;
            break;
        case KEY_STAR:
            if (zoomIn()) {
                check();
            }
            return;
        case KEY_POUND:
            if (zoomOut()) {
                check();
            }
            return;
        default:
            if ((browser.settings.flags & Settings.FAST_SCROLL) != 0) { 
                dx = stepX*scale;
                dy = stepY*scale;
                switch (getGameAction(keyCode)) { 
                case LEFT:
                    x -= dx;
                    break;
                case RIGHT:
                    x += dx;
                    break;
                case UP:
                    y -= dy;
                    break;
                case DOWN:
                    y += dy;
                    break;
                case FIRE:
                    if (!pointInfo()) { 
                        browser.error(Locale.current.getResource("No current point"), this);
                    }
                    return;
                case GAME_A:
                    if (zoomIn()) {
                        check();
                    }
                    return;
                case GAME_B:
                    if (zoomOut()) {
                        check();
                    }
                    return;
                default:
                    return;
                }
                check();
                return;
            } else { 
                switch (getGameAction(keyCode)) {
                case UP:
                    dy = -1;
                    break;
                case LEFT:
                    dx = -1;
                    break;
                case FIRE:
                    if (!pointInfo()) { 
                        browser.error(Locale.current.getResource("No current point"), this);
                    }
                    return;
                case RIGHT:
                    dx = 1;
                    break;
                case DOWN:
                    dy = 1;
                    break;
                }
            }
        }
        if (scale != 1) { 
            boolean redrawAll = false;
            focusX += dx*focusWidth*scale/zoom;
            focusY += dy*focusHeight*scale/zoom;
            if (focusX < x) { 
                redrawAll = true;
                x -= stepX*scale;
                focusX += stepX*scale;
                if (x < 0) { 
                    focusX += x;
                    x = 0;
                    if (focusX < 0) { 
                        focusX = 0;
                    }
                }
            }
            if (focusX - x > (screenWidth - focusWidth)*scale) { 
                redrawAll = true;
                x += stepX*scale;
                focusX -= stepX*scale;
                if (x + screenWidth*scale > browser.width) { 
                    focusX += x + screenWidth*scale - browser.width;
                    x = browser.width - screenWidth*scale;
                    if (focusX + focusWidth*scale > browser.width) { 
                        focusX = browser.width - focusWidth*scale;
                    }
                }
            }
            if (focusY < y) { 
                redrawAll = true;
                y -= stepY*scale;
                focusY += stepY*scale;
                if (y < 0) { 
                    focusY += y;
                    y = 0;
                    if (focusY < 0) { 
                        focusY = 0;
                    }
                }
            }
            if ((focusY - y) > (screenHeight - focusHeight)*scale) { 
                redrawAll = true;
                y += stepY*scale;
                focusY -= stepY*scale;
                if (y + screenHeight*scale > browser.height) { 
                    focusY += y + screenHeight*scale - browser.height;
                    y = browser.height - screenHeight*scale;
                    if (focusY + focusHeight*scale > browser.height) { 
                        focusY = browser.height - focusHeight*scale;
                    }
                }
            }
            if (redrawAll) { 
                repaint();
            } else { 
                repaint(((dx <= 0 ? focusX : focusX - focusWidth*scale/zoom) - x) / scale,
                        ((dy <= 0 ? focusY : focusY - focusHeight*scale/zoom) - y) / scale,
                        focusWidth + focusWidth/zoom + 2, 
                        focusHeight + focusHeight/zoom + 2);
            }
        } else { 
            x += dx*stepX;
            y += dy*stepY;
            check();
        }
    }


    private void check() { 
        if (x + screenWidth*scale > browser.width) { 
            x = browser.width - screenWidth*scale;
        }
        if (x < 0) { 
            x = 0;
        } 
        if (y + screenHeight*scale > browser.height) { 
            y = browser.height - screenHeight*scale;
        }
        if (y < 0) { 
            y = 0;
        }
        if (scale != 1) { 
            if (focusX < x || (focusX - x) > (screenWidth - focusWidth)*scale
                || focusY < y || (focusY - y) > (screenHeight - focusHeight)*scale)
            {
                focusX = x + focusX0;
                focusY = y + focusY0;
            }
            if (focusX + focusWidth*scale > browser.width) { 
                focusX = browser.width - focusWidth*scale;
            }
            if (focusY + focusHeight*scale > browser.height) { 
                focusY = browser.height - focusHeight*scale;
            }
        }
        repaint();
    }
}
